import polars as pl


string_type = pl.Expr | str


def negation(v: pl.NUMERIC_DTYPES) -> pl.Expr:
    """
    Apply negation to a Polars expression representing a numeric value.

    This function takes a numeric expression from the Polars library and
    returns its negated value. It is specifically designed for use with
    Polars expressions that contain numeric data types.

    Args:
        v (pl.NUMERIC_DTYPES): A Polars expression of a numeric data type.

    Returns:
        pl.Expr: A Polars expression representing the negated value of the
                 input expression.

    Example:
        >>> df = pl.DataFrame({'numbers': [1, -2, 3]})
        >>> df.select(negation(pl.col('numbers')))
        shape: (3, 1)
        ┌─────────┐
        │ numbers │
        │ ---     │
        │ i64     │
        ╞═════════╡
        │ -1      │
        ├─────────┤
        │ 2       │
        ├─────────┤
        │ -3      │
        └─────────┘
    """
    return pl.Expr.__neg__(v)


def negative() -> int:
    return -1




import polars as pl
from polars_expr_transformer.funcs import all_functions
from polars_expr_transformer.funcs.logic_functions import does_not_equal
from polars_expr_transformer.funcs.logic_functions import _in
operators = {  # get your data out of your code...
    "+": "pl.Expr.add",
    "-": "pl.Expr.sub",
    "*": "pl.Expr.mul",
    "%": "pl.Expr.mod",
    "/": "pl.Expr.truediv",
    "<": "pl.Expr.lt",
    "<=": "pl.Expr.le",
    ">": "pl.Expr.gt",
    ">=": "pl.Expr.ge",
    "==": "pl.Expr.eq",
    "=": "pl.Expr.eq",
    "&": "pl.Expr.and_",
    '|': "pl.Expr.or_",
    '!=': "does_not_equal",
    'and': "pl.Expr.and_",
    'or': "pl.Expr.or_",
    'in': "_in",
    'is_null': "pl.Expr.is_null",
}

aliases = {
    'not': '_not',
}


operators_mappings = {v: eval(v) for v in operators.values()}
all_split_vals = set(['(', ')', '$if$', '$endif$', '$else$', '$then$','$elseif$', ',', ''] + list(operators)+list(operators))
all_split_vals_reversed = [v[::-1] for v in all_split_vals]
funcs = {f'{k}': v for k,v in all_functions.items()}
funcs['pl.col'] = pl.col
funcs['pl.lit'] = pl.lit
funcs.update(operators_mappings)
for alias, ref in aliases.items():
    funcs[alias] = funcs[ref]


PRECEDENCE = {'+': 2, '-': 2, '*': 3, '/': 3, '|': -1, '&': -1, 'and': -1, 'or': -1, 'in': -1}


from typing import List, Union
from polars_expr_transformer.configs.settings import operators
from polars_expr_transformer.process.models import IfFunc, Classifier, Func
from polars_expr_transformer.process.tree import build_hierarchy


def reverse_dict(d: dict) -> dict:
    return {v: k for k, v in d.items()}


def inline_to_prefix_formula(inline_formula: List[Union[Classifier, IfFunc, Func]]):
    stack = []
    prefix_formula = []
    for i, token in enumerate(inline_formula):
        if not isinstance(token, Classifier):
            prefix_formula.append(token)
            continue
        if token.val == '(':
            stack.append(token)
        elif token.val == ')':
            while len(stack) > 0 and stack[-1].val != '(':
                prefix_formula.append(stack.pop())
            stack.pop()
        elif token.val in operators:
            while len(stack) > 0 and stack[-1].val != '(' and ((token.precedence if token.precedence is not None else 1) <= (stack[-1].precedence if stack[-1].precedence is not None else 0)):
                prefix_formula.append(stack.pop())
            stack.append(token)
        else:
            prefix_formula.append(token)
    while stack:
        prefix_formula.append(stack.pop())
    return prefix_formula[::-1]


def evaluate_prefix_formula(formula_tokens: List[Union[Classifier, IfFunc, Func]]):
    stack = []
    for token in reversed(formula_tokens):
        if token != '':
            if isinstance(token, Classifier):
                if token in reverse_dict(operators):
                    if len(stack) >= 2:
                        operand1 = stack.pop()
                        operand2 = stack.pop()
                        result = token, Classifier('('), operand2, Classifier(','),  operand1, Classifier(')')
                    elif len(stack) == 1:
                        operand1 = stack.pop()
                        result = token, Classifier('('), operand1, Classifier(')')
                    else:
                        result = token
                    stack.append(result)
                else:
                    stack.append(token)
            else:
                stack.append(token)
    return stack.pop()


def parse_formula(tokens:  List[Union[Classifier, IfFunc, Func]]):
    parsed_formula = []
    for val in tokens:
        if isinstance(val, Classifier):
            f = operators.get(val)
            if f is not None:
                parsed_formula.append(Classifier(f))
            else:
                parsed_formula.append(val)
        else:
            parsed_formula.append(val)
    return parsed_formula


def flatten_inline_formula(nested_classifier:  tuple[Classifier]) -> List[Classifier]:
    flat_result = []

    def flatten_result(vals: List[Classifier]):
        while len(vals)>0:
            current_val = vals.pop(0)
            if isinstance(current_val, (tuple, list)):
                flatten_result(list(current_val))
            else:
                flat_result.append(current_val)

    flatten_result(list(nested_classifier))
    return flat_result


def resolve_inline_formula(inline_formula_tokens: List[Classifier]):
    if any(c.val_type == 'operator' for c in inline_formula_tokens):
        prefixed_formula = inline_to_prefix_formula(inline_formula_tokens)
        parsed_prefixed_formula = parse_formula(prefixed_formula)
        evaluated_prefix_formula = evaluate_prefix_formula(parsed_prefixed_formula)
        return flatten_inline_formula(evaluated_prefix_formula)
    return inline_formula_tokens


def parse_inline_functions(_hierarchical_formula: Func):
    if any([a.val_type == 'operator' for a in _hierarchical_formula.args if isinstance(a, Classifier)]):
        prefixed_formula = inline_to_prefix_formula(_hierarchical_formula.args)
        parsed_prefixed_formula = parse_formula(prefixed_formula)
        evaluated_prefix_formula = evaluate_prefix_formula(parsed_prefixed_formula)
        flattened_prefix_formula = flatten_inline_formula(evaluated_prefix_formula)
        _hierarchical_formula.args = [build_hierarchy(flattened_prefix_formula)]
    else:
        for arg in _hierarchical_formula.args:
            if isinstance(arg, Func):
                parse_inline_functions(arg)
            elif isinstance(arg, IfFunc):
                for condition in arg.conditions:
                    parse_inline_functions(condition.condition)
                    parse_inline_functions(condition.val)
                parse_inline_functions(arg.else_val)




from typing import List
from polars_expr_transformer.process.models import Classifier


def replace_ambiguity_minus_sign(tokens: List[Classifier]) -> List[Classifier]:
    tokens_no_spaces = [t for t in tokens if t.val != '']
    if '-' not in [t.val for t in tokens]:
        return tokens
    i = 0
    new_tokens = []
    while i < len(tokens_no_spaces):
        current_token = tokens_no_spaces[i]
        if current_token.val == '-':
            if i == 0 or not tokens_no_spaces[i - 1].val.isnumeric():
                new_tokens.append(Classifier('__negative()'))
                new_tokens.append(Classifier('*'))
                i += 1
                continue
            elif tokens_no_spaces[i - 1].val.isnumeric():
                new_tokens.append(Classifier('+'))
                new_tokens.append(Classifier('__negative()'))
                new_tokens.append(Classifier('*'))
                i += 1
                continue
        new_tokens.append(tokens_no_spaces[i])
        i += 1
    return new_tokens


def standardize_quotes(tokens: List[str]):
    output_tokens = []
    for tok in tokens:
        if len(tok) > 1 and tok[0] == "'" and tok[-1] == "'":
            new_tok = '"' + tok[1:-1] + '"'
            output_tokens.append(new_tok)
        else:
            output_tokens.append(tok)
    return output_tokens


def standardize_tokens(tokens: List[str]) -> List[Classifier]:
    standardized_tokens = standardize_quotes(tokens)
    toks = [Classifier(val) for val in standardized_tokens]
    toks = [t for t in toks if t.val_type != 'empty']
    return replace_ambiguity_minus_sign(toks)



import re
from typing import Tuple, Dict, List
from copy import deepcopy



def find_last_occurrence(ls: List):
    for i in range(len(ls) - 1, -1, -1):
        if ls[i] == '(':
            return i

def reverse_dict(d: Dict) -> Dict:
    return {v: k for k, v in d.items()}


def contains_outside_of_quotes(func_string, *args):
    parts = re.split(r"""("[^"]*"|'[^']*')""", ' ' + func_string + ' ')
    vals = '~~~'.join((v.replace('(', '~~~').replace(')', '~~~').replace(' ', '~~~') for v in parts[::2]))
    return all(f'~~~{arg}~~~' in vals for arg in args)



def find_min_pos(fs: str, characters: Tuple):
    if any(character in fs for character in characters):
        return min(fs.find(c) for c in characters if c in fs)
    else:
        return 0


def test_if_position_is_in_quotes(func_string: str, pos: int):
    in_quotes = False
    for p, v in enumerate(func_string):
        if v in ('"', "'"):
            in_quotes = not in_quotes
            if p == pos:
                raise Exception
        if pos == p:
            return in_quotes


def test_if_func_ref(possible_func_ref: str) -> bool:
    return possible_func_ref.startswith('__f__') and possible_func_ref.endswith('__f__')


def contains_special_funcs(input_str: str) -> bool:
    if any(v in input_str for v in set(operators)):
        in_quotes = False
        data = []
        for w in input_str:
            if w in ('"', "'"):
                in_quotes = not in_quotes
            if w in operators and not in_quotes:
                return True
            data.append(w)
    return False


def replace_commas(input_str: str) -> str:
    in_quotes = False
    data = []
    for w in input_str:
        if w in ('"', "'"):
            in_quotes = not in_quotes
        if w == ',' and in_quotes:
            w = '%^%'
        data.append(w)
    return ''.join(data)


def get_last_value_of_generator(gen):
    last = 0
    for last in gen:  # This will end with 'last' being the last value or 0 if the generator is empty
        pass
    return last


def combine_functions_in_tokens(func_tokens: List[str]):
    first_close = next((i for i, t in enumerate(func_tokens) if t == ')'), 0)
    if first_close > 0:
        first_open = get_last_value_of_generator(i for i, t in enumerate(func_tokens[:first_close]) if t == '(')
        if first_open == 0:
            return func_tokens
        func_tokens = func_tokens[:first_open - 1] + [
            ''.join(func_tokens[first_open - 1:first_close + 1])] + func_tokens[first_close + 1:]
        return combine_functions_in_tokens(func_tokens)
    return func_tokens


def select_largest_non_overlapping(groups):
    # Sort by start point, then by reverse end point to get largest first
    sorted_groups = sorted(groups, key=lambda x: (x[0], -x[1]))
    result = []
    end = -float('inf')
    for group in sorted_groups:
        if group[0] > end:  # Check if non-overlapping
            result.append(group)
            end = group[1]  # Update last known end
    return result


def find_deepest_function(func_string: str) -> Tuple[Tuple[int, int], str]:
    def find_func_range(func_string: str):
        func_string_tokens = tokenize(func_string)
        has_first_close = ')' in func_string_tokens
        if has_first_close:
            first_close = func_string_tokens.index(')')
            last_open = find_last_occurrence(func_string_tokens[:first_close])
            len_of_func = len(''.join(func_string_tokens[last_open - 1:first_close + 1]))
            ref_start = len(''.join(func_string_tokens[:last_open - 1]))
            first_close = ref_start + len_of_func
            return ref_start, first_close
        else:
            return -1, -1

    def find_inline_func_range(func_string: str, check_for_equals: bool = True):
        # func_string = '-120=-121*0'
        tt = tokenize(func_string, False)
        if '=' in tt:
            tt = ''.join(['%~=~%' if t == '=' else t for t in tt]).split('%~=~%')
            results = [(find_inline_func_range(t, False), len(t) + i + 1) for i, t in enumerate(tt)]
            other_inlines = any(r[0][0] >= 0 and r[0][1] >= 0 for r in results)
            if other_inlines:
                pos = 0
                for r in results:
                    inline_func = r[0][0] >= 0 and r[0][1] >= 0
                    if inline_func:
                        return r[0][0] + pos, r[0][1] + pos
                    pos += r[1]

        func_string_tokens = combine_functions_in_tokens(tokenize(func_string, False))
        func_string_tokens = [t for t in func_string_tokens if t != '']
        inline_funcs = []
        for i, tok in enumerate(func_string_tokens):
            if tok in operators:
                inline_funcs.append(i)

        grouped_inline_funcs = []
        current_group = []
        for v in inline_funcs:
            if len(current_group) == 0:
                current_group = [v, v]
            elif v - current_group[1] > 2:
                grouped_inline_funcs.append(current_group)
                current_group = []
            else:
                current_group[1] = v
        if current_group != []:
            grouped_inline_funcs.append(current_group)
        unique_groups = select_largest_non_overlapping(grouped_inline_funcs)
        if len(unique_groups) > 0:
            first_group = unique_groups[0]
            start = sum(len(v) for v in func_string_tokens[:max(first_group[0] - 1, 0)])
            end = sum([len(v) for v in func_string_tokens[:first_group[-1] + 2]])
            return start, end

        else:
            return -1, -1

    def find_if_range(func_string: str):
        endif_index = func_string.find('$endif$')
        if endif_index != -1:
            matches = [match for match in re.finditer(r'\$if\$', func_string[:endif_index])]
            if len(matches) > 0:
                match = matches[-1]
            else:
                match = None
            if match:
                return match.start(), endif_index + len('$endif$')
        return -1, -1

    def test_inner(t1: Tuple[str, Tuple[int, int]], t2: Tuple[str, Tuple[int, int]]):
        # Calculate the range differences
        diff1 = t1[1][1] - t1[1][0]
        diff2 = t2[1][1] - t2[1][0]

        # If the ranges are non-overlapping and the second range starts after the first one ends,
        # return the first range as it starts first.
        if t1[1][1] < t2[1][0]:
            return t1
        if t2[1][1] < t1[1][0]:
            return t2
        if t1[1][1] < t2[1][1] and t1[1][0] > t2[1][0]:
            return t1
        if t2[1][1] < t1[1][1] and t2[1][0] > t1[1][0]:
            return t2
        if t1[0] == 'inline' and t2[0] in ('if', 'f'):
            return t2
        if t2[0] == 'inline' and t1[0] in ('if', 'f'):
            return t1
        return None

    def test_inners(*args):
        args = [a for a in args if sum(a[1]) > 0]
        if len(args) == 0:
            return None
        elif len(args) == 1:
            return args[0]
        while len(args) > 1:
            arg_1 = args.pop(0)
            arg_2 = args.pop(0)
            args.append(test_inner(arg_1, arg_2))
        return args[0]

    if_range = find_if_range(func_string)
    # print(f'deepest if range: {if_range}')
    func_range = find_func_range(func_string)
    # print(f'deepest func range: {func_range}')
    inline_range = find_inline_func_range(func_string)
    # print(f'deepest inline range: {inline_range}')
    inner_func = test_inners(('if', if_range), ('f', func_range), ('inline', inline_range))
    if inner_func is None:
        return None
    else:
        return inner_func[1], inner_func[0]


def replace_double_minus(toks: List[str]) -> List[str]:
    new_toks = []
    i = 0
    while i < len(toks):
        if i < len(toks) - 1 and toks[i] == '-' and toks[i + 1] == '-':
            new_toks.append('+')
            i += 2  # Skip the next index
        else:
            new_toks.append(toks[i])
            i += 1
    return new_toks


def replace_ambiguity_minus_sign(tokens: List[str]) -> List[str]:
    tokens_no_spaces = [t for t in tokens if t != '']
    # tokens_no_spaces = replace_double_minus(toks=tokens_no_spaces)
    if '-' not in tokens:
        return tokens
    i = 0
    new_tokens = []
    while i < len(tokens_no_spaces):
        current_token = tokens_no_spaces[i]
        if current_token == '-':
            if i == 0 or not tokens_no_spaces[i - 1].isnumeric():
                # new_tokens.append(f'pl.Expr.__neg__({tokens_no_spaces[i+1]})')
                new_tokens.append('__negative()')
                new_tokens.append('*')
                i += 1
                continue
            elif tokens_no_spaces[i - 1].isnumeric():
                new_tokens.append('+')
                new_tokens.append('__negative()')
                new_tokens.append('*')
                i += 1
                continue
        new_tokens.append(tokens_no_spaces[i])
        i += 1
    return new_tokens


def tokenize_after_standardization(formula):
    r = list(formula[::-1])
    output = []
    v = ''
    in_string = False  # Flag to track if we're inside a string literal
    in_brackets = False  # Flag to track if we're inside square brackets
    i = 0
    string_indicator = None
    while i < len(r):
        current_val = r[i]
        if current_val == string_indicator:
            string_indicator = None
            in_string = False
        elif current_val in ('"', "'") and string_indicator is None:
            in_string = True  # Toggle the in_string flag
            string_indicator = current_val
        elif current_val in ['[', ']']:
            in_brackets = not in_brackets  # Toggle the in_brackets flag
        elif current_val == '=' and not in_brackets and not in_string:
            if len(r) > i + 1:
                two_character_inline = r[i + 1] in ('<', '>', '=', '!')
                if two_character_inline:
                    current_val += r[i + 1]
                    i += 1
        elif current_val == '/' and not in_brackets:
            if len(r) > i + 2:
                three_characters_inline = r[i + 1:i + 3] == '/s/'
                if three_characters_inline:
                    current_val += r[i + 1:i + 3]
                    i += 2
                    in_string = not in_string

        if not in_string and not in_brackets and current_val[::-1] in all_split_vals:
            if i > 0:
                output.append(v)
            output.append(current_val)
            v = ''
        elif any([vv[::-1] in v for vv in all_split_vals if len(vv) > 1]):
            splitter = next(vv[::-1] for vv in all_split_vals if len(vv) > 1 and vv[::-1] in v)

            for toks in v.split(splitter):
                if len(toks) > 0:
                    output.append(toks)
            output.append(splitter)
            v = current_val
        else:
            v += current_val
        i += 1

    if any([vv[::-1] in v for vv in all_split_vals if len(vv) > 1]):
        splitter = next(vv[::-1] for vv in all_split_vals if len(vv) > 1 and vv[::-1] in v)

        for toks in v.split(splitter):
            if len(toks) > 0:
                output.append(toks)
        output.append(splitter)
    else:
        output.append(v)
    output = [''.join(reversed(v)) for v in output]
    output.reverse()
    return output


def transform_inline_formula_to_pl_formula(formula: str, convert_to_pl_cols: bool = False) -> str:
    precedence = {'+': 2, '-': 2, '*': 3, '/': 3, '|': -1, '&': -1}

    def get_val(v):
        if v in operators:
            return operators[v]
        return v

    def inline_to_prefix_formula(inline_formula):
        stack = []
        prefix_formula = []  # Define operator precedence
        for token in inline_formula:
            if token == '(':
                stack.append(token)
            elif token == ')':
                while stack and stack[-1] != '(':
                    prefix_formula.append(stack.pop())
                stack.pop()  # Remove '('
            elif token in operators:
                while stack and stack[-1] != '(' and precedence.get(token, 1) <= precedence.get(stack[-1], 0):
                    prefix_formula.append(stack.pop())
                stack.append(token)
            else:
                prefix_formula.append(token)

        while stack:
            prefix_formula.append(stack.pop())

        return prefix_formula[::-1]  # Reverse the formula to obtain prefix notation

    def evaluate_prefix_formula(formula):
        stack = []
        for token in reversed(formula):
            if token != '':
                if token in reverse_dict(operators):
                    if len(stack) >= 2:
                        operand1 = stack.pop()
                        operand2 = stack.pop()
                        result = token + '(' + operand2 + ',' + operand1 + ')'
                    elif len(stack) == 1:
                        operand1 = stack.pop()
                        result = token + '(' + operand1 + ')'
                    else:
                        result = token
                    stack.append(result)
                else:
                    stack.append(token)
        return stack.pop()

    tokens = tokenize(formula, convert_to_pl_cols)
    tokens = replace_ambiguity_minus_sign(tokens)
    prefix_formula = inline_to_prefix_formula(tokens)
    parsed_formula = [get_val(v) for v in prefix_formula]
    result = evaluate_prefix_formula(parsed_formula)
    return result



import pytest
from polars_expr_transformer.process.polars_expr_transformer import preprocess, simple_function_to_expr
import polars as pl

def test_simple_constant_expression():
    df = pl.from_dicts([{'a': 'edward', 'b': 'courtney'}, {'a': 'courtney', 'b': 'edward'}])
    result = df.select(simple_function_to_expr("'hallo world'"))
    expected = pl.DataFrame({'literal': ['hallo world']})
    assert result.equals(expected)

def test_combining_columns_expression():
    df = pl.from_dicts([{'a': 'edward', 'b': 'courtney'}, {'a': 'courtney', 'b': 'edward'}])
    result = df.select(simple_function_to_expr('[a] + " loves " + [b]').alias('literal'))
    expected = pl.DataFrame({'literal': ['edward loves courtney', 'courtney loves edward']})
    assert result.equals(expected)

def test_condition_expression():
    df = pl.from_dicts([{'a': 'edward', 'b': 'courtney'}, {'a': 'courtney', 'b': 'edward'}])
    result = df.select(simple_function_to_expr('"a" in [a]').alias('literal'))
    expected = pl.DataFrame({'literal': [True, False]})
    assert result.equals(expected)

def test_complex_conditional_expression():
    df = pl.from_dicts([{'a': 'edward', 'b': 'courtney'}, {'a': 'courtney', 'b': 'edward'}])
    result = df.select(simple_function_to_expr('concat("result:", if "a" in [a] then "A has been found" else "not found" endif)'))
    expected = pl.DataFrame({'literal': ['result:A has been found', 'result:not found']})
    assert result.equals(expected)

if __name__ == '__main__':
    pytest.main()



import re
from copy import deepcopy
from typing import List, Tuple



def replace_double_spaces(func_string: str):
    while '  ' in func_string:
        func_string = func_string.replace('  ', ' ')
    return func_string


def add_additions_outside_of_quotes(func_string: str, addition: str, *args):
    parts = re.split(r'("[^"]*"|\'[^\']*\')', func_string)
    parts[::2] = [replace_values(v, addition, *args) for v in parts[::2]]
    return ''.join(parts)


def replace_value_outside_of_quotes(func_string: str, val: str, replace: str):
    parts = re.split(r"""("[^"]*"|'[^']*')""", ' ' + func_string + ' ')
    parts[::2] = [v.replace(val, replace) for v in parts[::2]]
    return ''.join(parts)


def replace_values_outside_of_quotes(func_string: str, replacements: List[Tuple[str, str]]):
    for _old, _new in replacements:
        func_string = func_string.replace(_old, _new)
    return func_string


def replace_values(part_string: str, addition: str, *args):
    for arg in args:
        part_string = re.sub(rf'\b{arg}\b', f'{addition}{arg}{addition}', part_string)
    return part_string


def parse_pl_cols(func_string):
    func_op = []
    func_string = deepcopy(func_string)
    cur_string = func_string
    pos = 0
    inside_quotes = False
    quote_char = ''
    length = len(cur_string)

    while pos < length:
        char = cur_string[pos]
        if char in "\"'":
            if inside_quotes:
                if char == quote_char:
                    inside_quotes = False
                    quote_char = ''
            else:
                inside_quotes = True
                quote_char = char
        elif char == '[' and not inside_quotes:
            start = pos
            end = pos
            while end < length:
                end += 1
                if cur_string[end] in "\"'":
                    if inside_quotes:
                        if cur_string[end] == quote_char:
                            inside_quotes = False
                            quote_char = ''
                    else:
                        inside_quotes = True
                        quote_char = cur_string[end]
                elif cur_string[end] == ']' and not inside_quotes:
                    break

            if end < length and cur_string[end] == ']':
                val = cur_string[start + 1:end]
                if ',' not in val:
                    func_op.append((start + 1, end))
                pos = end
            else:
                break
        pos += 1

    col_rename = set((f'pl.col("{func_string[_s:_e]}")', func_string[_s - 1:_e + 1]) for _s, _e in func_op)
    for new_val, old_val in col_rename:
        func_string = func_string.replace(old_val, new_val)
    return func_string

def remove_unwanted_characters(func_string: str) -> str:
    parts = re.split(r"""("[^"]*"|'[^']*')""", func_string)
    parts[::2] = map(lambda s: "".join(s.split()), parts[::2])  # outside quotes
    return "".join(parts)


def preprocess(input_function: str):
    input_function = input_function.replace('\n' ,' ')
    input_function = replace_double_spaces(input_function)
    input_function = add_additions_outside_of_quotes(input_function, '$', 'if','else' ,'endif' ,'elseif' ,'then')
    input_function = replace_values_outside_of_quotes(input_function, replacements =[('$if$', '$if$('),
                                                                                      ('$else$', ')$else$('),
                                                                                      ('$endif$', ')$endif$'),
                                                                                      ('$elseif$', ')$elseif$('),
                                                                                      ('$then$', ')$then$(')])
    input_function = replace_value_outside_of_quotes(input_function, '==', '=')
    # input_function = standardize_quotes(input_function)
    input_function = parse_pl_cols(input_function)
    input_function = remove_unwanted_characters(input_function)
    return input_function



from typing import Optional, List
from polars_expr_transformer.process.models import Classifier, Func, IfFunc, TempFunc, ConditionVal
from copy import deepcopy


def handle_opening_bracket(current_func: Func, previous_val: Classifier) -> Func:
    if previous_val and previous_val.val_type == 'function':
        pass
    elif isinstance(current_func, IfFunc) and previous_val.val == '$if$':
        condition = Func(Classifier('pl.lit'))
        val = Func(Classifier('pl.lit'))
        condition_val = ConditionVal(condition=condition, val=val)
        else_val = Func(Classifier('pl.lit'))
        current_func.add_else_val(else_val)
        current_func.add_condition(condition_val)
        current_func = condition
    else:
        new_func = Func(Classifier('pl.lit'))
        current_func.add_arg(new_func)
        current_func = new_func
    return current_func

def handle_if(current_func: Func, current_val: Classifier) -> Func:
    new_func = IfFunc(current_val)
    current_func.add_arg(new_func)
    current_func = new_func
    return current_func

def handle_then(current_func: Func, current_val: Classifier, next_val: Optional[Classifier], pos: int) -> (Func, int):
    if isinstance(current_func, ConditionVal):
        current_func.func_ref = current_val
        current_func = current_func.val
        if next_val and next_val.val == '(':
            pos += 1
    return current_func, pos

def handle_else(current_func: Func, next_val: Optional[Classifier], pos: int) -> (Func, int):
    current_func = current_func.parent
    if isinstance(current_func, IfFunc):
        current_func = current_func.else_val
        if next_val and next_val.val == '(':
            pos += 1
    return current_func, pos

def handle_elseif(current_func: Func) -> Func:
    current_func = current_func.parent
    if isinstance(current_func, IfFunc):
        condition = Func(Classifier('pl.lit'))
        val = Func(Classifier('pl.lit'))
        condition_val = ConditionVal(condition=condition, val=val)
        current_func.add_condition(condition_val)
        current_func = condition
    return current_func

def handle_endif(current_func: Func) -> Func:
    if isinstance(current_func, IfFunc):
        current_func = current_func.parent
    else:
        raise Exception('Expected if')
    return current_func

def handle_closing_bracket(current_func: Func, main_func: Func) -> (Func, Func):
    if current_func.parent is None and current_func == main_func:
        new_main_func = TempFunc()
        new_main_func.add_arg(main_func)
        main_func = current_func = new_main_func
    elif current_func.parent is not None:
        current_func = current_func.parent
    else:
        raise Exception('Expected parent')
    return current_func, main_func

def handle_function(current_func: Func, current_val: Classifier) -> Func:
    new_function = Func(current_val)
    current_func.add_arg(new_function)
    current_func = new_function
    return current_func


def handle_literal(current_func: Func, current_val: Classifier):
    current_func.add_arg(current_val)


def build_hierarchy(tokens: List[Classifier]):
    # print_classifier(tokens)
    new_tokens = deepcopy(tokens)
    if new_tokens[0].val_type == 'function':
        main_func = Func(new_tokens.pop(0))
    else:
        main_func = Func(Classifier('pl.lit'))
    current_func = main_func
    pos = 0
    while pos < len(new_tokens):
        current_val = new_tokens[pos]
        previous_val = current_func.func_ref if pos < 1 else new_tokens[pos-1]
        next_val = new_tokens[pos+1] if len(new_tokens) > pos+1 else None
        if isinstance(current_val, Classifier):
            if current_val.val == '(':
                current_func = handle_opening_bracket(current_func, previous_val)
            elif current_val.val == '$if$':
                current_func = handle_if(current_func, current_val)
            elif current_val.val == '$then$':
                current_func, pos = handle_then(current_func, current_val, next_val, pos)
            elif current_val.val == '$else$':
                current_func, pos = handle_else(current_func, next_val, pos)
            elif current_val.val == '$elseif$':
                current_func = handle_elseif(current_func)
            elif current_val.val == '$endif$':
                current_func = handle_endif(current_func)
            elif current_val.val == ')':
                if next_val is None:
                    break
                current_func, main_func = handle_closing_bracket(current_func, main_func)
            elif current_val.val_type == 'function':
                current_func = handle_function(current_func, current_val)
            elif current_val.val_type in ('string', 'number', 'boolean', 'operator'):
                handle_literal(current_func, current_val)
            elif current_val.val == '__negative()':
                handle_literal(current_func, Classifier('-1'))
        else:
            handle_literal(current_func, current_val)
        pos += 1
    return main_func



import polars as pl
from typing import List, Tuple, Callable, Dict, Any
from polars_expr_transformer.funcs.utils import is_polars_expr, create_fix_col


string_type = pl.Expr | str


def concat(*columns) -> pl.Expr:
    """
    Concatenate the values of multiple columns into a single string.

    Parameters:
    - columns: List of columns to concatenate.

    Returns:
    - An expression representing the concatenated string.
    """
    columns = [create_fix_col(c) if not is_polars_expr(c) else c for c in columns]
    return pl.concat_str(columns)


def count_match(column: pl.Expr | str, value: str) -> pl.Expr:

    """
    Count the number of occurrences of a given value in a column.

    Parameters:
    - column: The column in which to count the occurrences.
    - value: The value to count.

    Returns:
    - An expression representing the number of occurrences.
    """
    if isinstance(column, pl.Expr):
        return column.str.count_match(value)
    return pl.lit(column).str.count_match(value)


def length(column: pl.Expr | str) -> pl.Expr:
    """
    Calculate the length of a given column or string.

    Parameters:
    - column: The column or string for which to calculate the length.

    Returns:
    - An expression representing the length.
    """
    if isinstance(column, str):
        return pl.lit(len(column))
    return column.str.lengths()


def uppercase(_v: pl.Expr | str) -> pl.Expr:
    """
    Convert the characters in a given column or string to uppercase.

    Parameters:
    - _v: The column or string to convert.

    Returns:
    - An expression representing the uppercase conversion.
    """
    if isinstance(_v, pl.Expr):
        return _v.str.to_uppercase()
    return pl.lit(_v.__str__().upper())


def to_string(_v: pl.Expr | str) -> pl.Expr:
    """
    Convert a given column or value to its string representation.

    Parameters:
    - _v: The column or value to convert.

    Returns:
    - An expression representing the string conversion.
    """
    if isinstance(_v, pl.Expr):
        return _v.cast(str)
    return pl.lit(_v.__str__())


def lowercase(_v: pl.Expr | str) -> pl.Expr:
    """
    Convert the characters in a given column or string to lowercase.

    Parameters:
    - _v: The column or string to convert.

    Returns:
    - An expression representing the lowercase conversion.
    """
    if isinstance(_v, pl.Expr):
        return _v.str.to_lowercase()
    return pl.lit(_v.__str__().lower())


def __left(row: Dict):
    v, l = row.values()
    if v is not None:
        return v[:l]
    return None

def left(column: string_type, length: pl.Expr | int) -> pl.Expr:
    """
    Extracts a substring from a column or string, starting from the beginning.

    Parameters:
    - column: The column or string from which to extract the substring.
    - length: The length of the substring to extract.

    Returns:
    - An expression representing the substring.
    """
    if is_polars_expr(column):
        if is_polars_expr(length):
            print(' pl.struct([column, length]).apply(lambda r: __left(r))')
            return pl.struct([column, length]).apply(lambda r: __left(r))
        else:
            print('column.str.slice(0, length)')
            return column.str.slice(0, length)
    elif is_polars_expr(length):
        print('pl.struct([length]).apply(lambda r: column[:list(r.values)[0]])')
        return pl.struct([length]).apply(lambda r: column[:list(r.values)[0]])
    else:
        print('pl.lit(column[:length])')
        return pl.lit(column[:length])


def __right(row: Dict):
    v, l = row.values()
    if v is not None:
        return v[-l:]


def right(column: string_type, length: pl.Expr | int) -> pl.Expr:
    """
    Extracts a substring from a column or string, starting from the end.

    Parameters:
    - column: The column or string from which to extract the substring.
    - length: The length of the substring to extract.

    Returns:
    - An expression representing the substring.
    """

    if is_polars_expr(column):
        if is_polars_expr(length):
            return pl.struct([column, length]).apply(__right)
        else:
            return column.str.slice(-length)
    elif is_polars_expr(length):
        return pl.struct([length]).apply(lambda r: column[-next(iter(r.values())):])
    else:
        return pl.lit(column[-length:])


def __apply_replace(row, replace_by=None):
    v = list(row.values())
    main_str = v[0]
    other_str = v[1] if len(v) > 1 else None
    replace_str = v[2] if len(v) > 2 else replace_by
    return main_str.replace(other_str, replace_str)


def replace(main: string_type, other: string_type, replace_by: string_type) -> pl.Expr:
    """
       Replaces occurrences of a substring within a main string with another string.

       Parameters:
       - main: The main string or expression where the replacement will occur.
       - other: The substring or expression that needs to be replaced.
       - replace_by: The string or expression that will replace the 'other' substring in 'main'.

       Returns:
       An expression representing the string after replacement.

       Example:
       Given main = "Hello, world!", other = "world", and replace_by = "there"
       The function will return "Hello, there!"
       """
    if not is_polars_expr(main):
        main = pl.lit(main)
    return main.str.replace_all(other, replace_by, literal=True).cast(pl.Utf8)


def to_date(s: Any, date_format: str = "%Y-%m-%d") -> pl.Expr:
    """
    Convert a string to a date.

    Parameters:
    - s (Any): The string to convert to a date. Can be a FlowFile expression or any other value.
    - format (str): The format of the date string. Default is "%Y-%m-%d".

    Returns:
    - pl.Expr: A FlowFile expression representing the converted date.

    Note: If `s` is not a FlowFile expression, it will be converted into one.
    """
    s = s if is_polars_expr(s) else create_fix_col(s)
    return s.str.to_date(date_format, strict=True)


def to_datetime(s: Any, date_format: str = "%Y-%m-%d %H:%M:%S") -> pl.Expr:
    """
    Convert a string to a datetime.

    Parameters:
    - s (Any): The string to convert to a datetime. Can be a FlowFile expression or any other value.
    - format (str): The format of the datetime string. Default is "%Y-%m-%d %H:%M:%S".

    Returns:
    - pl.Expr: A FlowFile expression representing the converted datetime.

    Note: If `s` is not a FlowFile expression, it will be converted into one.
    """
    s = s if is_polars_expr(s) else create_fix_col(s)
    return s.str.to_datetime(date_format, strict=False)



from typing import List, Union
from polars_expr_transformer.process.models import IfFunc, Func, TempFunc
from polars_expr_transformer.process.tree import build_hierarchy
from polars_expr_transformer.process.tokenize import tokenize
from polars_expr_transformer.process.standardize import standardize_tokens
from polars_expr_transformer.process.process_inline import parse_inline_functions
from polars_expr_transformer.process.preprocess import preprocess
import polars as pl


def finalize_hierarchy(hierarchical_formula: Union[Func | TempFunc | IfFunc]):
    if isinstance(hierarchical_formula, TempFunc) and len(hierarchical_formula.args) == 1:
        return hierarchical_formula.args[0]
    elif isinstance(hierarchical_formula, TempFunc) and len(hierarchical_formula.args) > 1:
        raise Exception('Expected only one argument')
    elif isinstance(hierarchical_formula, TempFunc) and len(hierarchical_formula.args) == 0:
        raise Exception('Expected at least one argument')
    return hierarchical_formula


def build_func(func_str: str = 'concat("1", "2")') -> Func:
    func_str_processed = preprocess(func_str)
    tokens = tokenize(func_str_processed)
    standardized_tokens = standardize_tokens(tokens)
    hierarchical_formula = build_hierarchy(standardized_tokens)
    parse_inline_functions(hierarchical_formula)
    return finalize_hierarchy(hierarchical_formula)


def simple_function_to_expr(func_str: str) -> pl.expr.Expr:
    func = build_func(func_str)
    return func.get_pl_func()



from polars_expr_transformer.configs.settings import PRECEDENCE
from typing import TypeAlias, Literal, List, Union, Optional, Any
from polars_expr_transformer.configs.settings import operators, funcs
from dataclasses import dataclass, field
import polars as pl


value_type: TypeAlias = Literal['string', 'number', 'boolean', 'operator', 'function', 'column', 'empty', 'case_when',
                                'prio', 'sep', 'special']


@dataclass
class Classifier:
    val: str
    val_type: value_type = None
    precedence: int = None
    parent: Optional[Union["Classifier", "Func"]] = field(repr=False, default=None)

    def __post_init__(self):
        self.val_type = self.get_val_type()
        self.precedence = self.get_precedence()

    def get_precedence(self):
        return PRECEDENCE.get(self.val)

    def get_val_type(self) -> value_type:
        if self.val.lower() in ['true', 'false']:
            return 'boolean'
        elif self.val in operators:
            return 'operator'
        elif self.val in ('(', ')'):
            return 'prio'
        elif self.val == '':
            return 'empty'
        elif self.val in funcs:
            return 'function'
        elif self.val in ('$if$', '$then$', '$else$', '$endif$'):
            return 'case_when'
        elif self.val.isdigit():
            return 'number'
        elif self.val == '__negative()':
            return 'special'
        elif self.val.isalpha():
            return 'string'
        elif self.val == ',':
            return 'sep'
        else:
            return 'string'

    def get_pl_func(self):
        if self.val_type == 'boolean':
            return True if self.val.lower() == 'true' else False
        elif self.val_type == 'function':
            return funcs[self.val]
        elif self.val_type in ('number', 'string'):
            return eval(self.val)
        elif self.val == '__negative()':
            return funcs['__negative']()
        else:
            raise Exception('Did not expect to get here')

    def get_repr(self):
        return str(self.val)

    def __eq__(self, other):
        return self.val == other

    def __hash__(self):
        return hash(self.val)


@dataclass
class Func:
    func_ref: Union[Classifier, "IfFunc"]
    args: List[Union["Func", Classifier, "IfFunc"]] = field(default_factory=list)
    parent: Optional["Func"] = field(repr=False, default=None)

    def add_arg(self, arg: Union["Func", Classifier, "IfFunc"]):
        self.args.append(arg)
        arg.parent = self

    def get_pl_func(self):
        if self.func_ref == 'pl.lit':
            if len(self.args)!=1:
                raise Exception('Expected must contain 1 argument not more not less')
            if isinstance(self.args[0].get_pl_func(), pl.expr.Expr):
                return self.args[0].get_pl_func()
            return funcs[self.func_ref.val](self.args[0].get_pl_func())
        args = [arg.get_pl_func() for arg in self.args]
        if any(isinstance(arg, pl.Expr) for arg in args) and any(not isinstance(arg, pl.Expr) for arg in args):
            standardized_args = []
            for arg in args:
                if not isinstance(arg, pl.Expr):
                    standardized_args.append(pl.lit(arg))
                else:
                    standardized_args.append(arg)
        else:
            standardized_args = args
        return funcs[self.func_ref.val](*standardized_args)


@dataclass
class ConditionVal:
    func_ref: Union[Classifier, "IfFunc", "Func"] = None
    condition: Func = None
    val: Func = None
    parent: "IfFunc" = field(repr=False, default=None)

    def __post_init__(self):
        if self.condition:
            self.condition.parent = self
        if self.val:
            self.val.parent = self

    def get_pl_func(self):
        return pl.when(self.condition.get_pl_func()).then(self.val.get_pl_func())

    def get_pl_condition(self):
        return self.condition.get_pl_func()

    def get_pl_val(self):
        return self.val.get_pl_func()


@dataclass
class IfFunc:
    func_ref: Union[Classifier]
    conditions: Optional[List[ConditionVal]] = field(default_factory=list)
    else_val: Optional[Func] = None
    parent: Optional[Func] = field(repr=False, default=None)

    def add_condition(self, condition: ConditionVal):
        self.conditions.append(condition)
        condition.parent = self

    def add_else_val(self, else_val: Func):
        self.else_val = else_val
        else_val.parent = self

    def get_pl_func(self):
        full_expr = None
        if len(self.conditions)==0:
            raise Exception('Expected at least one condition')
        for condition in self.conditions:
            if full_expr is None:
                full_expr = pl.when(condition.get_pl_condition()).then(condition.get_pl_val())
            else:
                full_expr = full_expr.when(condition.get_pl_condition()).then(condition.get_pl_val())
        return full_expr.otherwise(self.else_val.get_pl_func())


@dataclass
class TempFunc:
    args: List[Union["Func", Classifier, "IfFunc"]] = field(default_factory=list)

    def add_arg(self, arg: Union["Func", Classifier, "IfFunc"]):
        self.args.append(arg)
        arg.parent = self



import polars as pl
from typing import Any
from polars_expr_transformer.funcs.utils import is_polars_expr, create_fix_col, create_fix_date_col
from datetime import datetime

string_type = pl.Expr | str


def now() -> pl.Expr:
    """
    Get the current timestamp.

    Returns:
    - pl.Expr: A FlowFile expression representing the current timestamp.
    """
    return pl.lit(datetime.now())


def today() -> pl.Expr:
    """
    Get the current date.

    Returns:
    - pl.Expr: A FlowFile expression representing the current date.
    """
    return pl.lit(datetime.today())


def year(s: Any) -> pl.Expr:
    """
    Extract the year from a date or timestamp.

    Parameters:
    - s (Any): The date or timestamp to extract the year from. Can be a FlowFile expression or any other value.

    Returns:
    - pl.Expr: A FlowFile expression representing the year extracted from `s`.

    Note: If `s` is not a FlowFile expression, it will be converted into one.
    """
    s = s if is_polars_expr(s) else create_fix_date_col(s)
    return s.dt.year()


def month(s: Any) -> pl.Expr:
    """
    Extract the month from a date or timestamp.

    Parameters:
    - s (Any): The date or timestamp to extract the month from. Can be a FlowFile expression or any other value.

    Returns:
    - pl.Expr: A FlowFile expression representing the month extracted from `s`.

    Note: If `s` is not a FlowFile expression, it will be converted into one.
    """
    s = s if is_polars_expr(s) else create_fix_col(s).str.to_datetime()
    return s.dt.month()


def day(s: Any) -> pl.Expr:
    """
    Extract the day from a date or timestamp.

    Parameters:
    - s (Any): The date or timestamp to extract the day from. Can be a FlowFile expression or any other value.

    Returns:
    - pl.Expr: A FlowFile expression representing the day extracted from `s`.

    Note: If `s` is not a FlowFile expression, it will be converted into one.
    """
    s = s if is_polars_expr(s) else create_fix_date_col(s)
    return s.dt.day()


def hour(s: Any) -> pl.Expr:
    """
    Extract the hour from a timestamp.

    Parameters:
    - s (Any): The timestamp to extract the hour from. Can be a FlowFile expression or any other value.

    Returns:
    - pl.Expr: A FlowFile expression representing the hour extracted from `s`.

    Note: If `s` is not a FlowFile expression, it will be converted into one.
    """
    s = s if is_polars_expr(s) else create_fix_date_col(s)
    return s.dt.hour()


def minute(s: Any) -> pl.Expr:
    """
    Extract the minute from a timestamp.

    Parameters:
    - s (Any): The timestamp to extract the minute from. Can be a FlowFile expression or any other value.

    Returns:
    - pl.Expr: A FlowFile expression representing the minute extracted from `s`.

    Note: If `s` is not a FlowFile expression, it will be converted into one.
    """
    s = s if is_polars_expr(s) else create_fix_date_col(s)
    return s.dt.minute()


def second(s: Any) -> pl.Expr:
    """
    Extract the second from a timestamp.

    Parameters:
    - s (Any): The timestamp to extract the second from. Can be a FlowFile expression or any other value.

    Returns:
    - pl.Expr: A FlowFile expression representing the second extracted from `s`.

    Note: If `s` is not a FlowFile expression, it will be converted into one.
    """
    s = s if is_polars_expr(s) else create_fix_date_col(s)
    return s.dt.second()


def add_days(s: Any, days: Any) -> pl.Expr:
    """
    Add a specified number of days to a date or timestamp.

    Parameters:
    - s (Any): The date or timestamp to add days to. Can be a FlowFile expression or any other value.
    - days (int): The number of days to add.

    Returns:
    - pl.Expr: A FlowFile expression representing the result of adding `days` to `s`.

    Note: If `s` is not a FlowFile expression, it will be converted into one.
    """
    s = s if is_polars_expr(s) else create_fix_date_col(s)
    days = days if is_polars_expr(days) else create_fix_date_col(days)
    return s + pl.duration(days=days)


def add_hours(s: Any, hours: any) -> pl.Expr:
    """
    Add a specified number of hours to a timestamp.

    Parameters:
    - s (Any): The timestamp to add hours to. Can be a FlowFile expression or any other value.
    - hours (int): The number of hours to add.

    Returns:
    - pl.Expr: A FlowFile expression representing the result of adding `hours` to `s`.

    Note: If `s` is not a FlowFile expression, it will be converted into one.
    """
    s = s if is_polars_expr(s) else create_fix_date_col(s)
    hours = hours if is_polars_expr(hours) else create_fix_col(hours)
    return s + pl.duration(hours=hours)


def add_minutes(s: Any, minutes: Any) -> pl.Expr:
    """
    Add a specified number of minutes to a timestamp.

    Parameters:
    - s (Any): The timestamp to add minutes to. Can be a FlowFile expression or any other value.
    - minutes (int): The number of minutes to add.

    Returns:
    - pl.Expr: A FlowFile expression representing the result of adding `minutes` to `s`.

    Note: If `s` is not a FlowFile expression, it will be converted into one.
    """
    s = s if is_polars_expr(s) else create_fix_date_col(s)
    minutes = minutes if is_polars_expr(minutes) else create_fix_col(minutes)
    return s + pl.duration(minutes=minutes)


def add_seconds(s: Any, seconds: Any) -> pl.Expr:
    """
    Add a specified number of seconds to a timestamp.

    Parameters:
    - s (Any): The timestamp to add seconds to. Can be a FlowFile expression or any other value.
    - seconds (int): The number of seconds to add.

    Returns:
    - pl.Expr: A FlowFile expression representing the result of adding `seconds` to `s`.

    Note: If `s` is not a FlowFile expression, it will be converted into one.
    """
    s = s if is_polars_expr(s) else create_fix_date_col(s)
    seconds = seconds if is_polars_expr(seconds) else create_fix_col(seconds)
    return s + pl.duration(seconds=seconds)


def datetime_diff_seconds(s1: Any, s2: Any) -> pl.Expr:
    """
    Calculate the difference in seconds between two timestamps.

    Parameters:
    - s1 (Any): The first timestamp. Can be a FlowFile expression or any other value.
    - s2 (Any): The second timestamp. Can be a FlowFile expression or any other value.

    Returns:
    - pl.Expr: A FlowFile expression representing the difference in seconds between `s1` and `s2`.

    Note: If `s1` and `s2` are not FlowFile expressions, they will be converted into one.
    """
    s1 = s1 if is_polars_expr(s1) else create_fix_date_col(s1)
    s2 = s2 if is_polars_expr(s2) else create_fix_date_col(s2)
    return (s1 - s2).dt.total_seconds()


def datetime_diff_nanoseconds(s1: Any, s2: Any):
    """ Calculate the difference in days between two dates.
    Parameters:
    - s1 (Any): The first date. Can be a FlowFile expression or any other value.
    - s2 (Any): The second date. Can be a FlowFile expression or any other value.
    Returns:
    - pl.Expr: A FlowFile expression representing the difference in days between `s1` and `s2`.
    Note: If `s1` and `s2` are not FlowFile expressions, they will be converted into one.
    """
    s1 = s1 if is_polars_expr(s1) else create_fix_date_col(s1)
    s2 = s2 if is_polars_expr(s2) else create_fix_date_col(s2)
    return (s1 - s2).dt.total_nanoseconds()


def date_diff_days(s1: Any, s2: Any) -> pl.Expr:
    """
    Calculate the difference in days between two dates.

    Parameters:
    - s1 (Any): The first date. Can be a FlowFile expression or any other value.
    - s2 (Any): The second date. Can be a FlowFile expression or any other value.

    Returns:
    - pl.Expr: A FlowFile expression representing the difference in days between `s1` and `s2`.

    Note: If `s1` and `s2` are not FlowFile expressions, they will be converted into one.
    """
    s1 = s1 if is_polars_expr(s1) else create_fix_date_col(s1)
    s2 = s2 if is_polars_expr(s2) else create_fix_date_col(s2)
    return (s1 - s2).dt.total_days()


import os

from polars_expr_transformer.funcs import logic_functions, string_functions, math_functions, special_funcs, date_functions

all_functions = {}
all_functions.update(logic_functions.__dict__)
all_functions.update(string_functions.__dict__)
all_functions.update(math_functions.__dict__)
all_functions.update(special_funcs.__dict__)
all_functions.update(date_functions.__dict__)



from polars_expr_transformer.funcs.utils import is_polars_expr, create_fix_col
from typing import Any, Dict
import polars as pl


def equals(s: Any, t: Any) -> pl.Expr:
    """
    Check if two expressions or values are equal.

    Parameters:
    - s (Any): The first expression or value to compare. Can be a FlowFile expression or any other value.
    - t (Any): The second expression or value to compare. Can be a FlowFile expression or any other value.

    Returns:
    - pl.Expr: A FlowFile expression representing the equality of `s` and `t`.

    Note: If `s` or `t` is not a FlowFile expression, it will be converted into one.
    """
    s = s if is_polars_expr(s) else create_fix_col(s)
    t = t if is_polars_expr(t) else create_fix_col(t)
    return s.eq(t)


def is_empty(s: pl.Expr) -> pl.Expr:
    """
    Check if a given expression is empty.

    Parameters:
    - s (pl.Expr): The expression to check. Must be a FlowFile expression.

    Returns:
    - pl.Expr: A FlowFile expression representing whether `s` is empty.

    Note: If `s` is not a FlowFile expression, a ValueError will be raised.
    """
    return s.is_null()


def is_not_empty(s: pl.Expr) -> pl.Expr:
    """
    Check if a given expression is empty.

    Parameters:
    - s (pl.Expr): The expression to check. Must be a FlowFile expression.

    Returns:
    - pl.Expr: A FlowFile expression representing whether `s` is empty.

    Note: If `s` is not a FlowFile expression, a ValueError will be raised.
    """
    return s.is_not_null()


def does_not_equal(s: Any, t: Any):
    s = s if is_polars_expr(s) else create_fix_col(s)
    t = t if is_polars_expr(t) else create_fix_col(t)
    return pl.Expr.eq(s, t).is_not()


def _not(s: Any) -> pl.Expr:
    """
    Negate a given expression.

    Parameters:
    - s (pl.Expr): The expression to negate. Must be a FlowFile expression.

    Returns:
    - pl.Expr: A FlowFile expression representing the negation of `s`.

    Note: If `s` is not a FlowFile expression, a ValueError will be raised.
    """
    if not is_polars_expr(s):
        s = pl.lit(s)
    return pl.Expr.not_(s)


def is_string(val: Any) -> pl.Expr:
    """
    Check if a given expression or value is a string.

    Parameters:
    - val (Any): The expression or value to check. Can be a FlowFile expression or any other value.

    Returns:
    - pl.Expr: A FlowFile expression representing whether `val` is a string.

    Note: If `val` is a FlowFile expression, its dtype will be checked. Otherwise, Python's isinstance will be used.
    """
    if is_polars_expr(val):
        dtype = pl.select(val).dtypes[0]
        return pl.lit(dtype.is_(pl.Utf8))
    return pl.lit(isinstance(val, str))


def contains(base: pl.Expr | str, pattern: pl.Expr | str) -> pl.Expr:
    """
    Check if a pattern is contained within a base string or expression.

    Parameters:
    - base (Union[Expr, str]): The base string or expression where the pattern is to be searched.
    - pattern (Union[Expr, str]): The pattern string or expression to search for within the base.

    Returns:
    - Expr: An expression indicating whether the pattern is contained within the base.

    Note: 
    - If both base and pattern are expressions, a custom function is used for performance optimization.
    - If one of them is a string and the other is an expression, built-in string matching is used.
    - If both are strings, Python's native 'in' operator is used and the result is wrapped as an expression.
    """
    def _contain(_row: Dict):
        base_val, pattern_val = _row.values()
        return pattern_val in base_val

    if isinstance(base, pl.Expr):
        if isinstance(pattern, pl.Expr):
            return pl.struct([base, pattern]).apply(lambda r: _contain(r), return_dtype=pl.Boolean)
        else:
            return base.str.contains(pattern)
    else:
        if isinstance(pattern, pl.Expr):
            return pl.struct([pattern]).apply(lambda x: next(iter(x.values())) in base)
        else:
            return pl.lit(pattern in base)


def _in(pattern: pl.Expr | str, base: pl.Expr | str) -> pl.Expr:
    return contains(base, pattern)


from pydantic import BaseModel
from typing import Optional, List


class ExpressionRef(BaseModel):
    name: str
    doc: Optional[str]


class ExpressionsOverview(BaseModel):
    expression_type: str
    expressions: List[ExpressionRef]



import os
import importlib.util
import inspect
from schemas import ExpressionRef, ExpressionsOverview
from typing import List, Optional


_available_expressions: Optional[ExpressionsOverview] = None


def get_formula_scripts():
    try:
        current_file_ref = os.path.abspath(__file__)
        current_file_path = os.path.dirname(current_file_ref)
    except:
        current_file_path = os.path.sep.join(('backend','flowfile','rowFuncs'))
    script_file_path = os.path.join(current_file_path, 'funcs')
    python_files = ((f[:-3], os.path.join(script_file_path,f)) for f in os.listdir(script_file_path) if '.py')
    filtered_files = (f for f in python_files if f[0][:min(2,len(f))] != '__' and f[0] != 'utils')
    return filtered_files


def get_module_members(module_name: str, path: str):
    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    module_members = inspect.getmembers(module)
    module_function_names = [ExpressionRef(name=name, doc=obj.__doc__) for name, obj in module_members
                             if inspect.isfunction(obj) and obj.__module__ == module.__name__
                             and name[:min(len(name),2)]!='__']
    return module_function_names


def get_expression_overview() -> List[ExpressionsOverview]:
    global _available_expressions
    if _available_expressions is None:
        scripts = get_formula_scripts()
        module_members = [ExpressionsOverview(expression_type = module_name,
                                              expressions = get_module_members(module_name, path))
                          for module_name, path in scripts]
        _available_expressions = module_members
    return _available_expressions


def get_all_expressions() -> List[str]:
    expression_overview = get_expression_overview()
    return [expression.name for eo in expression_overview for expression in eo.expressions]







from polars_expr_transformer.configs.settings import all_split_vals, all_functions


def tokenize(formula: str):
    r = list(formula[::-1])
    output = []
    v = ''
    in_string = False  # Flag to track if we're inside a string literal
    in_brackets = False  # Flag to track if we're inside square brackets
    i = 0
    string_indicator = None
    while i < len(r):
        current_val = r[i]
        if current_val == string_indicator:
            output.append(v+current_val)
            v = ''
            string_indicator = None
            in_string = False
            i += 1
            continue
        elif current_val in ('"', "'") and string_indicator is None:
            in_string = True  # Toggle the in_string flag
            string_indicator = current_val
        elif current_val in ['[', ']']:
            in_brackets = not in_brackets  # Toggle the in_brackets flag
        elif current_val == '=' and not in_brackets and not in_string:
            if len(r) > i + 1:
                two_character_inline = r[i + 1] in ('<', '>', '=', '!')
                if two_character_inline:
                    current_val += r[i + 1]
                    i += 1

        if not in_string and not in_brackets and current_val[::-1] in all_split_vals:
            if i > 0:
                output.append(v)
            output.append(current_val)
            v = ''
        elif any([vv[::-1] in v+current_val for vv in all_split_vals if len(vv) > 1]) and not in_string:
            splitter = next(vv[::-1] for vv in all_split_vals if len(vv) > 1 and vv[::-1] in v+current_val)
            if splitter:
                # check for longer possiblities
                longer_options = [f for f in all_functions.keys() if (v+current_val)[::-1] in f]
                if len(longer_options)>0:
                    temp_i, temp_v = i, v
                    while len([f for f in all_functions.keys() if temp_v[::-1] in f])>0 and temp_i<len(r):
                        temp_v += r[temp_i]
                        temp_i += 1
                    other_split = next((f for f in all_functions.keys() if temp_v[::-1] == f), None)
                    next_value = r[temp_i+1] if temp_i<len(r) else None
                    if next_value in (None, ' ', '(') and other_split is not None:
                        output.append(temp_v)
                        v = next_value
                        i = temp_i + 1
                        continue
            for toks in (v+current_val).split(splitter):
                if len(toks) > 0:
                    output.append(toks)
            output.append(splitter)
            v = ''
        else:
            v += current_val
        i += 1

    if v is not None and any([vv[::-1] in v for vv in all_split_vals if len(vv) > 1]):
        splitter = next(vv[::-1] for vv in all_split_vals if len(vv) > 1 and vv[::-1] in v)

        for toks in v.split(splitter):
            if len(toks) > 0:
                output.append(toks)
        output.append(splitter)
    elif v is not None:
        output.append(v)
    output = [''.join(reversed(v)) for v in output]
    output.reverse()
    return output



import polars as pl
from typing import Any
import os


def is_polars_expr(v: Any) -> bool:
    return isinstance(v, pl.Expr)


def create_fix_col(val: Any) -> pl.Expr:
    return pl.lit(val)


def create_fix_date_col(s: Any) -> pl.Expr:
    return pl.col(s).str.to_datetime()



from setuptools import setup, find_packages
from pathlib import Path

# Read the contents of your README file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

# Function to parse requirements.txt
def parse_requirements(filename):
    """Load requirements from a pip requirements file."""
    with open(filename, 'r') as file:
        lines = file.readlines()
    return [line.strip() for line in lines if line and not line.startswith("#")]

setup(
    name='polars_expr_transformer',  # Package name
    version='0.1.0',  # Package version
    description='Transform string-based expressions into Polars DataFrame operations',  # Short description
    long_description=long_description,
    long_description_content_type='text/markdown',  # Use markdown format for long description
    author='Edward van Eehoud',  # Replace with your name
    author_email='evaneechoudl@gmail.com',  # Replace with your email
    url='https://github.com/yourusername/polars_expr_transformer',  # Replace with the URL to your package repository
    packages=find_packages(include=['polars_expr_transformer', 'polars_expr_transformer.*']),  # Automatically find package directories
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',  # Specify the minimum Python version required
    install_requires=parse_requirements('requirements.txt'),  # Parse requirements from requirements.txt
    extras_require={
        'dev': [
            'check-manifest',
            'flake8',
        ],
        'test': [
            'coverage',
            'pytest',
        ],
    },
    include_package_data=True,  # Include package data specified in MANIFEST.in
)